@@ -3,6 +3,7 @@
 
 import logging
 import os
+import warnings
 
 from sdsstools import get_config, get_logger, get_package_version
 from sdsstools.configuration import __ENVVARS__
@@ -22,6 +23,15 @@ __ENVVARS__["OBSERVATORY"] = "?"
 config = get_config(NAME)
 
 
+# If we are not in debug mode, remove some possible warnings.
+if config["debug"] is False:
+    warnings.filterwarnings(
+        "ignore",
+        message=".+was never awaited.+",
+        category=RuntimeWarning,
+    )
+
+
 def start_file_loggers(start_log=True, start_can=True):
 
     if "files" in config and "log_dir" in config["files"]:
