@@ -25,6 +25,7 @@ from sdsstools.daemonizer import DaemonGroup
 from jaeger import can_log, config, log
 from jaeger.commands.bootloader import load_firmware
 from jaeger.commands.calibration import calibrate_positioner
+from jaeger.exceptions import JaegerError, JaegerUserWarning
 from jaeger.fps import FPS
 from jaeger.testing import VirtualFPS
 
@@ -36,7 +37,7 @@ def shutdown(sign):
     """Shuts down the FPS and stops the positioners in case of a signal interrupt."""
 
     if __FPS__ is not None:
-        __FPS__.send_command("STOP_TRAJECTORY", positioner_id=0, synchronous=True)
+        __FPS__.send_command("STOP_TRAJECTORY", positioner_ids=0, now=True)
         log.error(f"stopping positioners and cancelling due to {sign.name}")
         sys.exit(0)
     else:
@@ -61,17 +62,18 @@ def cli_coro(f):
 class FPSWrapper(object):
     """A helper to store FPS initialisation parameters."""
 
-    def __init__(self, profile, layout, ieb=None, danger=None, initialise=True):
+    def __init__(self, profile, ieb=None, initialise=True, npositioners=10):
 
         self.profile = profile
         if self.profile in ["test", "virtual"]:
             self.profile = "virtual"
 
-        self.layout = layout
         self.ieb = ieb
-        self.danger = danger
         self.initialise = initialise
 
+        self.vpositioners = []
+        self.npositioners = npositioners
+
         self.fps = None
 
     async def __aenter__(self):
@@ -81,14 +83,11 @@ class FPSWrapper(object):
         # If profile is test we start a VirtualFPS first so that it can respond
         # to the FPS class.
         if self.profile == "virtual":
-            self.fps = VirtualFPS(layout=self.layout)
+            self.fps = VirtualFPS()
+            for pid in range(self.npositioners):
+                self.fps.add_virtual_positioner(pid + 1)
         else:
-            self.fps = FPS(
-                can_profile=self.profile,
-                layout=self.layout,
-                ieb=self.ieb,
-                engineering_mode=self.danger,
-            )
+            self.fps = FPS(can=self.profile, ieb=self.ieb)
 
         __FPS__ = self.fps
 
@@ -98,7 +97,11 @@ class FPSWrapper(object):
         return self.fps
 
     async def __aexit__(self, *excinfo):
-        await self.fps.shutdown()
+        try:
+            assert self.fps
+            await self.fps.shutdown()
+        except JaegerError as err:
+            warnings.warn(f"Failed shutting down FPS: {err}", JaegerUserWarning)
 
 
 pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)
@@ -119,10 +122,16 @@ pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)
     help="The bus interface profile.",
 )
 @click.option(
-    "-l",
-    "--layout",
-    type=str,
-    help="The FPS layout.",
+    "--virtual",
+    is_flag=True,
+    help="Runs a virtual FPS with virtual positioners. Same as --profile=virtual.",
+)
+@click.option(
+    "-n",
+    "--npositioners",
+    type=int,
+    default=10,
+    help="How many virtual positioners must be connected to the virtual FPS.",
 )
 @click.option(
     "-v",
@@ -141,11 +150,6 @@ pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)
     default=None,
     help="Does not connect to the IEB.",
 )
-@click.option(
-    "--danger",
-    is_flag=True,
-    help="Enables engineering mode. Most safety checks will be disabled.",
-)
 @click.option(
     "-s",
     "--sextant",
@@ -154,7 +158,17 @@ pass_fps = click.make_pass_decorator(FPSWrapper, ensure=True)
     "Modifies the internal configuration file.",
 )
 @click.pass_context
-def jaeger(ctx, config_file, layout, profile, verbose, quiet, ieb, danger, sextant):
+def jaeger(
+    ctx,
+    config_file,
+    profile,
+    verbose,
+    quiet,
+    ieb,
+    sextant,
+    virtual,
+    npositioners,
+):
     """CLI for the SDSS-V focal plane system.
 
     If called without subcommand starts the actor.
@@ -164,27 +178,40 @@ def jaeger(ctx, config_file, layout, profile, verbose, quiet, ieb, danger, sexta
     if verbose > 0 and quiet:
         raise click.UsageError("--quiet and --verbose are mutually exclusive.")
 
+    if config_file:
+        config.load(config_file)
+
+    actor_config = config.get("actor", {})
+
     if verbose == 1:
         log.sh.setLevel(logging.INFO)
+        actor_config["verbose"] = logging.INFO
     elif verbose == 2:
         log.sh.setLevel(logging.DEBUG)
+        actor_config["verbose"] = logging.DEBUG
     elif verbose >= 3:
         log.sh.setLevel(logging.DEBUG)
         can_log.sh.setLevel(logging.DEBUG)
+        actor_config["verbose"] = logging.DEBUG
 
     if quiet:
-        log.sh.propagate = False  # type: ignore
+        log.handlers.remove(log.sh)
         warnings.simplefilter("ignore")
-
-    if config_file:
-        config.load(config_file)
+        actor_config["verbose"] = 100
 
     if sextant:
         sextant_file = os.path.join(os.path.dirname(__file__), "etc/sextant.yaml")
         config["files"]["ieb_config"] = sextant_file
         log.debug(f"Using internal IEB sextant onfiguration file {sextant_file}.")
 
-    ctx.obj = FPSWrapper(profile, layout, ieb, danger)
+    if virtual is True:
+        profile = "virtual"
+
+    ctx.obj = FPSWrapper(
+        profile,
+        ieb=ieb,
+        npositioners=npositioners,
+    )
 
 
 @jaeger.group(cls=DaemonGroup, prog="jaeger_actor", workdir=os.getcwd())
@@ -224,6 +251,18 @@ async def actor(fps_maker, no_tron):
     nargs=1,
     type=click.Path(exists=True),
 )
+@click.argument(
+    "SEXTANTS",
+    type=int,
+    nargs=-1,
+    required=False,
+)
+@click.option(
+    "-y",
+    "--yes",
+    is_flag=True,
+    help="Do not ask for confirmation.",
+)
 @click.option(
     "-f",
     "--force",
@@ -231,22 +270,22 @@ async def actor(fps_maker, no_tron):
     help="Forces skipping of invalid positioners",
 )
 @click.option(
-    "-s",
+    "-p",
     "--positioners",
     type=str,
     help="Comma-separated positioners to upgrade",
 )
 @click.option(
-    "-c",
-    "--cycle",
+    "-n",
+    "--no-cycle",
     is_flag=True,
-    help="Power cycle positioners before upgrade to set them to bootloader mode. "
-    "If --sextant not used, cycles all sextants",
+    help="Does not power cycle positioners before upgrading each sextant",
 )
 @click.option(
-    "--sextant",
-    type=int,
-    help="The sextant to power cycle",
+    "-o",
+    "--all-on",
+    is_flag=True,
+    help="Powers on all sextants after a successful upgrade.",
 )
 @pass_fps
 @cli_coro
@@ -255,36 +294,57 @@ async def upgrade_firmware(
     firmware_file,
     force,
     positioners,
-    cycle,
-    sextant,
+    no_cycle,
+    sextants,
+    all_on,
+    yes,
 ):
     """Upgrades the firmaware."""
 
     if positioners is not None:
         positioners = [int(positioner.strip()) for positioner in positioners.split(",")]
 
+    fps_maker.initialise = False
+
+    sextants = sextants or (1, 2, 3, 4, 5, 6)
+
+    if not yes:
+        click.confirm(
+            f"Upgrade firmware for sextant(s) {', '.join(map(str, sextants))}?",
+            default=False,
+            abort=True,
+        )
+
     async with fps_maker as fps:
 
-        if fps.ieb and cycle:
-            log.info("power cycling positioners")
-            if sextant:
-                devs = [f"PS{sextant}"]
-            else:
-                devs = [f"PS{s}" for s in range(1, 7)]
+        for sextant in sextants:
+            log.info(f"Upgrading firmware on sextant {sextant}.")
 
-            await asyncio.gather(*[fps.ieb.get_device(dev).open() for dev in devs])
-            await asyncio.sleep(5)
-            await asyncio.gather(*[fps.ieb.get_device(dev).close() for dev in devs])
+            if fps.ieb and no_cycle is False:
+                log.info("power cycling positioners")
+                dev = "PS" + str(sextant)
+                await asyncio.gather(
+                    *[fps.ieb.get_device(f"PS{sextant}").open() for sextant in sextants]
+                )
+                await asyncio.sleep(5)
+                await fps.ieb.get_device(dev).close()
+                await asyncio.sleep(3)
+
+            await fps.initialise(start_pollers=False)
+
+            await load_firmware(
+                fps,
+                firmware_file,
+                positioners=positioners,
+                force=force,
+                show_progressbar=True,
+            )
+
+    if all_on is True:
+        log.info("Powering on sextants.")
+        for sextant in sextants:
+            await fps.ieb.get_device(f"PS{sextant}").close()
             await asyncio.sleep(3)
-            await fps.initialise()
-
-        await load_firmware(
-            fps,
-            firmware_file,
-            positioners=positioners,
-            force=force,
-            show_progressbar=True,
-        )
 
 
 @jaeger.command()
@@ -508,3 +568,7 @@ async def home(fps_maker, positioner_id):
         await asyncio.gather(*[positioner.home() for positioner in valid_positioners])
 
     return
+
+
+if __name__ == "__main__":
+    jaeger()
