@@ -11,12 +11,14 @@ import math
 
 import click
 
+from clu.command import Command
 from clu.parsers.click import pass_args
 
 from jaeger.fps import FPS
 from jaeger.ieb import IEB
 from jaeger.testing import VirtualFPS
 
+from .. import JaegerActor
 from . import jaeger_parser
 
 
@@ -70,7 +72,7 @@ def ieb(command, fps):
 async def status(command, fps):
     """Outputs the status of the devices."""
 
-    categories = fps.ieb.get_categories()
+    categories = sorted(fps.ieb.get_categories())
 
     status_data = {}
     for category in categories:
@@ -86,7 +88,7 @@ async def status(command, fps):
 @click.option(
     "--on/--off",
     default=None,
-    help="the value of the device. If not provided, switches the current status.",
+    help="the desired state of the device. If not passed, switches the current state.",
 )
 @click.option(
     "--cycle",
@@ -108,6 +110,9 @@ async def switch(command, fps, device, on, cycle):
     except ValueError:
         return command.fail(error=f"cannot find device {device!r}.")
 
+    if dev_name == "SYNC":
+        return command.fail(error="the SYNC line cannot be switched manually.")
+
     if device_obj.module.mode not in ["holding_register", "coil"]:
         return command.fail(error=f"{dev_name!r} is not an output.")
 
@@ -130,8 +135,6 @@ async def switch(command, fps, device, on, cycle):
     except Exception:
         return command.fail(error=f"failed to set status of device {dev_name!r}.")
 
-    command.debug(message={category: await _get_category_data(command, category)})
-
     if cycle:
         command.write("d", text="waiting 1 second before powering up.")
         await asyncio.sleep(1)
@@ -241,10 +244,15 @@ def power(command, fps):
 
 
 @power.command()
-async def on(command, fps):
+async def on(command, fps: FPS):
     """Powers on all the FPS IEB components."""
 
-    ieb: IEB = fps.ieb
+    command.info(text="Turning pollers off.")
+    await fps.pollers.stop()
+
+    ieb = fps.ieb
+    if not isinstance(ieb, IEB) or ieb.disabled:
+        return command.fail(error="IEB is not conencted or is disabled.")
 
     # Sequence of relays to power on. Tuples indicate relays that can be powered
     # on concurrently.
@@ -281,7 +289,12 @@ async def on(command, fps):
 async def off(command, fps, nucs):
     """Powers off all the FPS IEB components."""
 
-    ieb: IEB = fps.ieb
+    command.info(text="Turning pollers off.")
+    await fps.pollers.stop()
+
+    ieb = fps.ieb
+    if not isinstance(ieb, IEB) or ieb.disabled:
+        return command.fail(error="IEB is not conencted or is disabled.")
 
     # Sequence of relays to power off. Tuples indicate relays that can be powered
     # off concurrently.
@@ -302,14 +315,7 @@ async def off(command, fps, nucs):
     ]
 
     if nucs:
-        off_seq += [
-            "NUC1",
-            "NUC2",
-            "NUC3",
-            "NUC4",
-            "NUC5",
-            "NUC6",
-        ]
+        off_seq += ["NUC1", "NUC2", "NUC3", "NUC4", "NUC5", "NUC6"]
 
     if not (await _power_sequence(command, ieb, off_seq, mode="off")):
         return
@@ -325,6 +331,9 @@ async def fbi(command, fps: FPS, device_name: str, value: float):
 
     raw_value = 32 * int(1023 * (value / 100))
 
+    if not isinstance(fps.ieb, IEB) or fps.ieb.disabled:
+        return command.fail(error="IEB is not conencted or is disabled.")
+
     try:
         device = fps.ieb.get_device(device_name)
     except ValueError:
@@ -337,3 +346,67 @@ async def fbi(command, fps: FPS, device_name: str, value: float):
 
     await device.write(raw_value)
     return command.finish()
+
+
+@ieb.command()
+@click.option("-v", "--verbose", is_flag=True, help="Shows extra information.")
+async def info(command: Command[JaegerActor], fps: FPS, verbose=False):
+    """Shows information about the IEB layout in a human-readable format."""
+
+    ieb = fps.ieb
+
+    if not isinstance(ieb, IEB):
+        return command.fail(error="IEB is not conencted or is disabled.")
+
+    modules = sorted(ieb.modules.keys())
+    categories = set()
+
+    command.info(text="Modules:")
+
+    for module_name in modules:
+        module = ieb.modules[module_name]
+
+        command.info(text=f"  {module.name}:")
+
+        if module.description != "":
+            command.info(text=f"    description: {module.description}")
+
+        if verbose:
+            if module.model:
+                command.info(text=f"    model: {module.model}")
+            if module.mode:
+                command.info(text=f"    mode: {module.mode}")
+            if module.channels:
+                command.info(text=f"    channels: {module.channels}")
+
+        command.info(text="    Devices:")
+
+        for dev in ieb[module_name].devices.values():
+            command.info(text=f"      {dev.name}:")
+            if dev.category != "":
+                command.info(text=f"        category: {dev.category}")
+                categories.add(dev.category)
+            if dev.description != "":
+                command.info(text=f"        description: {dev.description}")
+
+            if verbose:
+                if dev.address:
+                    command.info(text=f"        address: {dev.address}")
+                if dev.channel:
+                    command.info(text=f"        channel: {dev.channel}")
+                if dev.__type__:
+                    command.info(text=f"        type: {dev.__type__}")
+                    if dev.__type__ == "relay" and dev.relay_type:
+                        command.info(text=f"        relay type: {dev.relay_type}")
+
+    if len(categories) > 0 and command.actor and command.actor.model:
+
+        command.info(text="")
+        command.info(text="Keywords:")
+
+        for category in sorted(categories):
+            cat_data = command.actor.model.schema["properties"][category]
+            items = [item["title"] for item in cat_data["items"]]
+            command.info(text=f"  {category}=[{', '.join(items)}]")
+
+    return command.finish()
