@@ -230,11 +230,6 @@ async def status(command, fps, positioner_id, full):
 
     command.info(locked=fps.locked)
 
-    if fps.engineering_mode:
-        command.warning(engineering_mode=True)
-    else:
-        command.info(engineering_mode=False)
-
     command.info(n_positioners=len(fps.positioners))
 
     for pid in sorted(positioner_ids):
@@ -300,7 +295,7 @@ async def current(command, fps, positioner_id, alpha, beta, all):
         return command.fail(error="FPS is moving. Cannot send set current.")
 
     commands = [
-        fps.send_command(SetCurrent(positioner_id=pid, alpha=alpha, beta=beta))
+        fps.send_command(SetCurrent(pid, alpha=alpha, beta=beta))
         for pid in positioner_id
     ]
     await asyncio.gather(*commands)
@@ -395,7 +390,7 @@ async def on(command, fps, positioner_id):
         return
 
     command.debug("Turning hall sensors ON")
-    await fps.send_to_all("HALL_ON", positioners=positioner_id)
+    await fps.send_command("HALL_ON", positioner_ids=positioner_id)
 
     command.debug("Waiting 5 seconds ...")
     await asyncio.sleep(5)
@@ -415,6 +410,59 @@ async def off(command, fps, positioner_id):
         return
 
     command.debug("Turning hall sensors OFF")
-    await fps.send_to_all("HALL_OFF", positioners=positioner_id)
+    await fps.send_command("HALL_OFF", positioner_ids=positioner_id)
 
     command.finish()
+
+
+@jaeger_parser.group()
+def led():
+    """Turns the positioner LED on/off."""
+
+    pass
+
+
+@led.command(name="on")
+@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
+async def led_on(command, fps, positioner_id):
+    """Turns the LED on."""
+
+    if positioner_id is None:
+        positioner_id = list(fps.positioners.keys())
+
+    if not check_positioners(positioner_id, command, fps, initialised=False):
+        return
+
+    command.debug("Turning LED on")
+    await fps.send_command("SWITCH_LED_ON", positioner_ids=positioner_id)
+
+    command.finish()
+
+
+@led.command(name="off")
+@click.argument("POSITIONER-ID", type=int, nargs=-1, required=False)
+async def led_off(command, fps, positioner_id):
+    """Turns the LED off."""
+
+    if positioner_id is None:
+        positioner_id = list(fps.positioners.keys())
+
+    if not check_positioners(positioner_id, command, fps, initialised=False):
+        return
+
+    command.debug("Turning LED off")
+    await fps.send_command("SWITCH_LED_OFF", positioner_ids=positioner_id)
+
+    command.finish()
+
+
+@jaeger_parser.command()
+async def reload(command, fps):
+    """Reinitialise the FPS."""
+
+    try:
+        await fps.initialise(start_pollers=fps.pollers.running)
+    except BaseException as err:
+        return command.fail(error=f"Initialisation failed: {err}")
+
+    return command.finish(text="FPS was reinitialised.")
