@@ -14,31 +14,44 @@ import collections
 import pprint
 import re
 import socket
-import time
-import warnings
+from dataclasses import dataclass, field
 
-from typing import Any, Callable, Generic, List, Optional, Type, TypeVar
-
-import can
-from can.interfaces.slcan import slcanBus
-from can.interfaces.socketcan import SocketcanBus
-from can.interfaces.virtual import VirtualBus
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    Generic,
+    List,
+    Optional,
+    Type,
+    TypeVar,
+    cast,
+)
 
 import jaeger
-import jaeger.interfaces.cannet
-import jaeger.utils
 from jaeger import can_log, config, log, start_file_loggers
-from jaeger.commands import Command, CommandID, Message, StopTrajectory
-from jaeger.exceptions import JaegerUserWarning
-from jaeger.interfaces.cannet import CANNetBus
+from jaeger.commands import Command, CommandID
+from jaeger.exceptions import JaegerCANError
+from jaeger.interfaces import BusABC, CANNetBus, Message, Notifier, VirtualBus
 from jaeger.maskbits import CommandStatus
-from jaeger.utils import Poller
+from jaeger.utils import Poller, parse_identifier
 
 
-__all__ = ["JaegerCAN", "CANnetInterface", "JaegerReaderCallback", "INTERFACES"]
+try:
+    from can.interfaces.slcan import slcanBus  # type: ignore
+    from can.interfaces.socketcan import SocketcanBus  # type: ignore
+except ImportError:
+    SocketcanBus = None
+    slcanBus = None
+
+if TYPE_CHECKING:
+    from .fps import FPS
 
 
-LOG_HEADER = "({cmd.command_id.name}, {cmd.positioner_id}, {cmd.command_uid}):"
+__all__ = ["JaegerCAN", "CANnetInterface", "INTERFACES"]
+
+
+LOG_HEADER = "({cmd.command_id.name}, {cmd.command_uid}):"
 
 #: Accepted CAN interfaces and whether they are multibus.
 INTERFACES = {
@@ -49,282 +62,198 @@ INTERFACES = {
 }
 
 
-class JaegerReaderCallback(can.Listener):
-    """A message reader that triggers a callback on message received.
-
-    Parameters
-    ----------
-    callback
-        The function to run when a new message is received.
-    loop
-        If an asyncio event loop, the callback will be called with
-        ``call_soon``, otherwise it will be called immediately.
-
-    """
-
-    def __init__(
-        self,
-        callback: Callable[[can.Message], Any],
-        loop: Optional[asyncio.AbstractEventLoop] = None,
-    ):
-
-        self.callback = callback
-        self.loop = loop
-
-    def on_message_received(self, msg):
-        """Calls the callback with the received message."""
-
-        if self.loop:
-            self.loop.call_soon(self.callback, msg)
-        else:
-            self.callback(msg)
-
-
-Bus_co = TypeVar("Bus_co", bound="can.BusABC", covariant=True)
+Bus_co = TypeVar("Bus_co", bound="BusABC")
+T = TypeVar("T", bound="JaegerCAN")
 
 
+@dataclass
 class JaegerCAN(Generic[Bus_co]):
     """A CAN interface with a command queue and reply handling.
 
     Provides support for multi-channel CAN networks, with each channel being
-    able to host more than one bus. In general, a new instance of `.JaegerCAN`
-    is create via the `~.JaegerCAN.from_profile` classmethod.
+    able to host more than one bus. The recommended way to instantiate a new
+    `.JaegerCAN` object is using the `.create` classmethod ::
+
+        can = await JaegerCAN.create(...)
+
+    which is equivalent to ::
+
+        can = JaegerCAN(...)
+        await can.start()
 
     Parameters
     ----------
     interface_type
-        One of `~jaeger.can.INTERFACES`. Defines the
-        `python-can <https://python-can.readthedocs.io/en/stable/>`_ interface
-        to use.
+        One of `~jaeger.can.INTERFACES`.
     channels
         A list of channels to be used to instantiate the interfaces.
     fps
         The focal plane system.
-    args,kwargs
-        Arguments and keyword arguments to pass to the interfaces when
+    interface_args
+        Keyword arguments to pass to the interfaces when
         initialising it (e.g., port, baudrate, etc).
 
-    Attributes
-    ----------
-    command_queue : asyncio.Queue
-        Queue of messages to be sent to the bus. The messages are sent as
-        soon as the bus has finished processing any commands with the same
-        ``command_id`` and ``positioner_id``.
-    listener : JaegerReaderCallback
-        A `.JaegerReaderCallback` instance that runs a callback when
-        a new message is received from the bus.
-    multibus : bool
-        Whether the interfaces are multibus.
-    notifier : can.Notifier
-        A `can.Notifier` instance that processes messages from the list
-        of buses, asynchronously.
-
     """
 
-    def __init__(
-        self,
-        interface_type: str,
-        channels: list | tuple,
-        *args,
-        fps: Optional[jaeger.FPS] = None,
-        **kwargs,
-    ):
+    interface_type: str
+    channels: list | tuple
+    fps: Optional[jaeger.fps.FPS] = None
+    interface_args: Dict[str, Any] = field(default_factory=dict)
+
+    def __post_init__(self):
+
+        if self.interface_type not in INTERFACES:
+            raise ValueError(f"Invalid interface {self.interface_type}.")
 
         # Start can file logger
-        start_file_loggers(start_log=False, start_can=True)
+        if config["debug"] is True:
+            start_file_loggers(start_log=False, start_can=True)
 
-        self.loop = asyncio.get_event_loop()
-
-        assert interface_type in INTERFACES, f"invalid interface {interface_type}."
-        self.interface_type = interface_type
-
-        InterfaceClass: Type[Bus_co] = INTERFACES[interface_type]["class"]
-
-        self.multibus = INTERFACES[interface_type]["multibus"]
-
-        if not isinstance(channels, (list, tuple)):
-            channels = [channels]
-
-        self.fps = fps
-
-        #: list: A list of `python-can`_ interfaces, one for each of the ``channels``.
+        # List of interfaces for each channel.
         self.interfaces: List[Bus_co] = []
-        for channel in channels:
-            log.info(
-                f"creating interface {interface_type}, "
-                f"channel={channel!r}, args={args}, kwargs={kwargs}."
-            )
+
+        self.multibus: bool = False
+
+        self._started: bool = False
+
+        # Currently running commands.
+        self.running_commands: Dict[int, Command] = {}
+
+        self.command_queue: asyncio.Queue[Command] | None = None
+        self._command_queue_task: asyncio.Task | None = None
+
+        self.notifier: Notifier | None = None
+
+    async def start(self: T) -> T:
+
+        itype = self.interface_type
+
+        InterfaceClass: Type[Bus_co] = INTERFACES[itype]["class"]
+        self.multibus = INTERFACES[itype]["multibus"]
+
+        if not isinstance(self.channels, (list, tuple)):
+            self.channels = [self.channels]
+
+        for channel in self.channels:
+            iargs = "".join([f" {k}={repr(v)}" for k, v in self.interface_args.items()])
+            log.debug(f"creating interface {itype}, channel={channel!r}{iargs}.")
             try:
-                self.interfaces.append(InterfaceClass(channel, *args, **kwargs))
+                interface = InterfaceClass(channel, **self.interface_args)
+                result = await interface.open()
+                if result is False:
+                    raise ConnectionError()
+                self.interfaces.append(interface)
             except ConnectionResetError:
                 log.error(
-                    f"connection to {interface_type}:{channel} failed. "
+                    f"connection to {itype}:{channel} failed. "
                     "Possibly another instance is connected to the device."
                 )
-            except (socket.timeout, ConnectionRefusedError, OSError):
-                log.error(
-                    f"connection to {interface_type}:{channel} failed. "
-                    "The device is not responding."
-                )
+            except (socket.timeout, ConnectionError, ConnectionRefusedError, OSError):
+                log.error(f"connection to {itype}:{channel} failed.")
             except Exception as ee:
-                raise ee.__class__(
-                    f"connection to {interface_type}:{channel} failed: {ee}."
-                )
+                raise ee.__class__(f"connection to {itype}:{channel} failed: {ee}.")
 
-        if len(self.interfaces) == 0:
-            warnings.warn("cannot connect to any interface.", JaegerUserWarning)
+        self.command_queue = asyncio.Queue()
+        self._command_queue_task = asyncio.create_task(self._process_command_queue())
 
-        self._start_notifier()
-
-        #: list: Currently running commands.
-        self.running_commands = []
-
-        self.command_queue: asyncio.Queue[Command] = asyncio.Queue()
-        self._command_queue_task = self.loop.create_task(self._process_queue())
-
-    def _start_notifier(self):
-        """Starts the listener and notifiers."""
-
-        self.listener = JaegerReaderCallback(self._process_reply, loop=self.loop)
-        self.notifier = can.notifier.Notifier(
-            self.interfaces,
-            [self.listener],
-            loop=self.loop,
+        self.notifier = Notifier(
+            listeners=[self._process_reply_queue],
+            buses=self.interfaces,
         )
 
-        log.debug("started JaegerReaderCallback listener and notifiers")
+        self._started = True
 
-    def _process_reply(self, msg: can.Message):
-        """Processes replies from the bus."""
+        return self
 
-        positioner_id, command_id, reply_uid, __ = jaeger.utils.parse_identifier(
-            msg.arbitration_id
-        )
+    @classmethod
+    async def create(
+        cls,
+        profile: Optional[str] = None,
+        fps: FPS = None,
+        interface_type: Optional[str] = None,
+        channels: list | tuple = [],
+        interface_args: Dict[str, Any] = {},
+    ) -> JaegerCAN:
+        """Create and initialise a new bus interface from a configuration profile.
 
-        if command_id == CommandID.COLLISION_DETECTED:
+        This is the preferred method to initialise a new `.JaegerCAN` instance and is
+        equivalent to calling ``JaegerCAN`` and then `~.JaegerCAN.start`.
 
-            log.error(
-                f"a collision was detected in positioner {positioner_id}. "
-                "Sending STOP_TRAJECTORIES and locking the FPS."
-            )
+        Parameters
+        ----------
+        profile
+            The name of the profile that defines the bus interface, or `None`
+            to use the default configuration.
+        fps
+            The focal plane system.
+        interface_type
+            One of `~jaeger.can.INTERFACES`. Cannot be used with ``profile``.
+        channels
+            A list of channels to be used to instantiate the interfaces.
+        interface_args
+            Keyword arguments to pass to the interfaces when
+            initialising it (e.g., port, baudrate, etc).
 
-            # Manually send the stop trajectory to be sure it has
-            # priority over other messages. No need to do it if the FPS
-            # has been locked, which means that we have already stopped
-            # trajectories.
+        """
 
-            if self.fps and self.fps.locked:
-                return
+        if profile is not None or (profile is None and interface_type is None):
+            if "profiles" not in config:
+                raise ValueError("No 'interfaces' section in the configuration file.")
 
-            stop_trajectory_command = StopTrajectory(positioner_id=0)
-            self.send_to_interfaces(stop_trajectory_command.get_messages()[0])
+            if profile is None:
+                if "default" not in config["profiles"]:
+                    raise ValueError("Default interface not defined in configuration.")
+                profile = config["profiles"]["default"]
 
-            # Now lock the FPS. No need to abort trajectories because we just did.
-            if self.fps:
-                self.loop.create_task(self.fps.lock(stop_trajectories=False))
-                return
+            if profile not in config["profiles"]:
+                raise ValueError(f"Invalid interface profile {profile}")
 
-        if command_id == 0:
-            can_log.warning(
-                "invalid command with command_id=0, "
-                f"arbitration_id={msg.arbitration_id} received. "
-                "Ignoring it."
-            )
-            return
+            config_data = config["profiles"][profile].copy()
 
-        command_id_flag = CommandID(command_id)
+            interface_type = config_data.pop("interface")
+            if interface_type not in INTERFACES:
+                raise ValueError(f"invalid interface {interface_type}")
 
-        # Remove done running command. Leave the failed and cancelled ones in
-        # the list for 60 seconds to be able to catch delayed replies. We
-        # also sort them so that the most recent commands are found first.
-        # This is important for timed out broadcast still in the list while
-        # another instance of the same command is running. We want replies to
-        # be sent to the running command first.
-        self.running_commands = sorted(
-            [
-                rcmd
-                for rcmd in self.running_commands
-                if not rcmd.status == rcmd.status.DONE
-                and (time.time() - rcmd.start_time) < 60
-            ],
-            key=lambda cmd: cmd.start_time,
-            reverse=True,
-        )
-
-        found_cmd = False
-        for r_cmd in self.running_commands:
-            if r_cmd.command_id == command_id:
-                if (reply_uid == 0 and r_cmd.positioner_id == 0) or (
-                    reply_uid in r_cmd.message_uids
-                    and positioner_id == r_cmd.positioner_id
-                ):
-                    found_cmd = r_cmd
-                    break
-
-        if found_cmd:
-            can_log.debug(
-                f"({command_id_flag.name}, "
-                f"{positioner_id}, {found_cmd.command_uid}): "
-                f"queuing reply UID={reply_uid} "
-                f"to command {found_cmd.command_uid}."
-            )
-            found_cmd.reply_queue.put_nowait(msg)
-        else:
-            can_log.debug(
-                f"({command_id_flag.name}, {positioner_id}): "
-                f"cannot find running command for reply UID={reply_uid}."
-            )
-
-    def send_to_interfaces(
-        self,
-        message: Message,
-        interfaces: Optional[List[Bus_co]] = None,
-        bus: Optional[Any] = None,
-    ):
-        """Sends the message to the appropriate interface and bus."""
-
-        log_header = (
-            f"({message.command.command_id.name}, "
-            f"{message.command.positioner_id} "
-            f"{message.command.command_uid!s}): "
-        )
-
-        if len(self.interfaces) == 1 and not self.multibus:
-            data_hex = binascii.hexlify(message.data).decode()
-            can_log.debug(
-                log_header + "sending message with "
-                f"arbitration_id={message.arbitration_id}, "
-                f"UID={message.uid}, "
-                f"and data={data_hex!r} to interface."
-            )
-            self.interfaces[0].send(message)
-            return
-
-        # If not interface, send the message to all interfaces.
-        if interfaces is None:
-            interfaces = self.interfaces
-
-        for iface in interfaces:
-
-            iface_idx = self.interfaces.index(iface)
-            data_hex = binascii.hexlify(message.data).decode()
-            can_log.debug(
-                log_header + "sending message with "
-                f"arbitration_id={message.arbitration_id}, "
-                f"UID={message.uid}, "
-                f"and data={data_hex!r} to "
-                f"interface {iface_idx}, "
-                f"bus={0 if not bus else bus!r}."
-            )
-
-            if bus:
-                iface.send(message, bus=bus)  # type: ignore
+            if "channel" in config_data:
+                channels = [config_data.pop("channel")]
+            elif "channels" in config_data:
+                channels = config_data.pop("channels")
+                assert isinstance(channels, (list, tuple)), "channels must be a list"
             else:
-                iface.send(message)
+                raise KeyError("channel or channels key not found.")
 
-    async def _process_queue(self):
+            interface_args = config_data
+
+        elif profile is not None and interface_type is not None:
+            raise JaegerCANError("profile and interface_type are mutually exclusive.")
+
+        if interface_type == "cannet":
+            cls = CANnetInterface
+
+        assert interface_type, "interface_type not set. This should not have happened."
+
+        instance = cls(
+            interface_type,
+            channels=channels,
+            fps=fps,
+            interface_args=interface_args,
+        )
+
+        await instance.start()
+
+        return instance
+
+    def refresh_running_commands(self):
+        """Clears completed commands."""
+
+        rc = self.running_commands
+        self.running_commands = {key: cmd for key, cmd in rc.items() if not cmd.done()}
+
+    async def _process_command_queue(self):
         """Processes messages in the command queue."""
 
+        assert self.command_queue
+
         while True:
 
             cmd = await self.command_queue.get()
@@ -340,18 +269,85 @@ class JaegerCAN(Generic[Bus_co]):
                     cmd.cancel()
                 continue
 
-            can_log.debug(log_header + " sending messages to CAN bus.")
-
-            cmd.status = CommandStatus.RUNNING
-
             try:
-                self._send_messages(cmd)
-                self.running_commands.append(cmd)
+                self.send_messages(cmd)
             except jaeger.JaegerError as ee:
                 can_log.error(f"found error while getting messages: {ee}")
                 continue
 
-    def _send_messages(self, cmd: Command):
+    async def _process_reply_queue(self, msg: Message):
+        """Processes one reply message."""
+
+        positioner_id, command_id, reply_uid, __ = parse_identifier(msg.arbitration_id)
+
+        if command_id == CommandID.COLLISION_DETECTED:
+
+            log.error(
+                f"A collision was detected in positioner {positioner_id}. "
+                "Sending STOP_TRAJECTORIES and locking the FPS."
+            )
+
+            # Manually send the stop trajectory to be sure it has
+            # priority over other messages. No need to do it if the FPS
+            # has been locked, which means that we have already stopped
+            # trajectories.
+
+            if self.fps and self.fps.locked:
+                return
+
+            assert self.fps
+            await self.fps.stop_trajectory()
+
+            # Now lock the FPS. No need to abort trajectories because we just did.
+            if self.fps:
+                await self.fps.lock(stop_trajectories=False)
+                return
+
+        if command_id == 0:
+            can_log.warning(
+                "invalid command with command_id=0, "
+                f"arbitration_id={msg.arbitration_id} received. "
+                "Ignoring it."
+            )
+            return
+
+        command_id_flag = CommandID(command_id)
+
+        self.refresh_running_commands()
+
+        cmd_key = (positioner_id << 25) + (command_id << 15) + reply_uid
+
+        if cmd_key in self.running_commands:
+            running_cmd = self.running_commands[cmd_key]
+        elif (command_id << 15) + reply_uid in self.running_commands:
+            # Checks if the reply corresponds to a broadcast.
+            cmd_key = (command_id << 15) + reply_uid
+            running_cmd = self.running_commands[cmd_key]
+        else:
+            can_log.debug(
+                f"[{command_id_flag.name}, {positioner_id}]: "
+                f"cannot find a matching running command."
+            )
+            return
+
+        if not running_cmd.is_broadcast:
+            if reply_uid not in running_cmd.message_uids:
+                can_log.debug(
+                    f"[{command_id_flag.name}, {positioner_id}]: "
+                    f"matching command does not contain reply UID={reply_uid}."
+                )
+                return
+
+        can_log.debug(
+            f"[{command_id_flag.name}, "
+            f"{positioner_id}, {running_cmd.command_uid}]: "
+            f"queuing reply UID={reply_uid} "
+            f"to command {running_cmd.command_uid}."
+        )
+
+        running_cmd.process_reply(msg)
+
+    def send_messages(self, cmd: Command):
         """Sends messages to the interface.
 
         This method exists separate from _process_queue so that it can be used
@@ -360,68 +356,67 @@ class JaegerCAN(Generic[Bus_co]):
         """
 
         log_header = LOG_HEADER.format(cmd=cmd)
+
+        if cmd.status != CommandStatus.READY:
+            if cmd.status != CommandStatus.CANCELLED:
+                can_log.error(
+                    f"{log_header} command is not ready "
+                    f"(status={cmd.status.name!r})"
+                )
+                cmd.cancel()
+            return
+
+        can_log.debug(
+            f"{log_header} sending command {cmd.command_uid} "
+            f"to positioners {cmd.positioner_ids!r}."
+        )
+        can_log.debug(log_header + " sending messages to CAN bus.")
+
         messages = cmd.get_messages()
 
         for message in messages:
 
-            # Get the interface and bus to which to send the message
-            interfaces = getattr(cmd, "_interfaces", None)
-            bus = getattr(cmd, "_bus", None)
-
             if cmd.status.failed:
                 can_log.debug(
-                    log_header
-                    + " not sending more messages "
-                    + "since this command has failed."
+                    f"{log_header} not sending more messages "
+                    "since this command has failed."
                 )
-                self.running_commands.remove(cmd)
                 break
 
-            self.send_to_interfaces(message, interfaces=interfaces, bus=bus)
+            cmd_key = message.positioner_id << 25
+            cmd_key += message.command.command_id << 15
+            cmd_key += message.uid
 
-    @classmethod
-    def from_profile(cls, profile: Optional[str] = None, **kwargs) -> JaegerCAN:
-        """Creates a new bus interface from a configuration profile.
+            self.running_commands[cmd_key] = message.command
 
-        Parameters
-        ----------
-        profile
-            The name of the profile that defines the bus interface, or `None`
-            to use the default configuration.
+            # Get the interface and buses to which to send this command.
+            interfaces = self.interfaces
+            bus = None
+            is_multibus = self.multibus or len(self.interfaces) > 1
+            if is_multibus and message.positioner_id != 0:
+                if self.fps and message.positioner_id in self.fps.positioner_to_bus:
+                    interface, bus = self.fps.positioner_to_bus[message.positioner_id]
+                    interfaces = [cast(Bus_co, interface)]
 
-        """
+            for iface in interfaces:
 
-        assert (
-            "profiles" in config
-        ), "configuration file does not have an interfaces section."
+                iface_idx = self.interfaces.index(iface)
+                data_hex = binascii.hexlify(message.data).decode()
+                can_log.debug(
+                    log_header + "sending message with "
+                    f"arbitration_id={message.arbitration_id}, "
+                    f"UID={message.uid}, "
+                    f"and data={data_hex!r} to "
+                    f"interface {iface_idx}, "
+                    f"bus={0 if not bus else bus!r}."
+                )
 
-        if profile is None:
-            assert (
-                "default" in config["profiles"]
-            ), "default interface not set in configuration."
-            profile = config["profiles"]["default"]
+                if bus:
+                    iface.send(message, bus=bus)  # type: ignore
+                else:
+                    iface.send(message)
 
-        if profile not in config["profiles"]:
-            raise ValueError(f"invalid interface profile {profile}")
-
-        config_data = config["profiles"][profile].copy()
-
-        interface = config_data.pop("interface")
-        if interface not in INTERFACES:
-            raise ValueError(f"invalid interface {interface}")
-
-        if "channel" in config_data:
-            channels = [config_data.pop("channel")]
-        elif "channels" in config_data:
-            channels = config_data.pop("channels")
-            assert isinstance(channels, (list, tuple)), "channels must be a list"
-        else:
-            raise KeyError("channel or channels key not found.")
-
-        if interface == "cannet":
-            return CANnetInterface(interface, channels, **kwargs, **config_data)
-
-        return cls(interface, channels, **kwargs, **config_data)
+        cmd.status = CommandStatus.RUNNING
 
     @staticmethod
     def print_profiles() -> List[str]:
@@ -469,6 +464,7 @@ CANNET_ERRORS = {
 }
 
 
+@dataclass
 class CANnetInterface(JaegerCAN[CANNetBus]):
     r"""An interface class specifically for the CAN\@net 200/420 device.
 
@@ -477,12 +473,20 @@ class CANnetInterface(JaegerCAN[CANNetBus]):
 
     """
 
-    def __init__(self, *args, **kwargs):
-
-        super().__init__(*args, **kwargs)
+    status_interval: float = 5
 
+    def __post_init__(self):
+        super().__post_init__()
         self._device_status = collections.defaultdict(dict)
 
+    async def start(self):
+        r"""Starts CAN\@net connection."""
+
+        await super().start()
+
+        if len(self.interfaces) == 0:
+            return
+
         # Get ID and version of interfaces. No need to ask for this in each
         # status poll.
         for interface in self.interfaces:
@@ -494,17 +498,17 @@ class CANnetInterface(JaegerCAN[CANNetBus]):
         self.device_status_poller = Poller(
             "cannet_device",
             self._get_device_status,
-            delay=5,
+            delay=self.status_interval,
         )
         self.device_status_poller.start()
 
-    def _process_reply(self, msg: can.Message):
+    async def _process_reply_queue(self, msg: Message):
         """Processes a message checking first if it comes from the device."""
 
         if msg.arbitration_id == 0:
             return self.handle_device_message(msg)
 
-        super()._process_reply(msg)
+        await super()._process_reply_queue(msg)
 
     @property
     def device_status(self):
@@ -515,7 +519,7 @@ class CANnetInterface(JaegerCAN[CANNetBus]):
 
         return self._device_status
 
-    def handle_device_message(self, msg: can.Message):
+    def handle_device_message(self, msg: Message):
         """Handles a reply from the device (i.e., not from the CAN network)."""
 
         device_status = self._device_status
