@@ -22,6 +22,7 @@ from sdsstools import read_yaml_file
 from jaeger import config, log, maskbits
 from jaeger.commands import Command, CommandID
 from jaeger.exceptions import FPSLockedError, TrajectoryError
+from jaeger.ieb import IEB
 from jaeger.utils import int_to_bytes
 
 
@@ -52,7 +53,7 @@ async def send_trajectory(
     fps: FPS,
     trajectories: str | pathlib.Path | TrajectoryDataType,
     use_sync_line=True,
-):
+) -> Trajectory:
     """Sends a set of trajectories to the positioners.
 
     This is a low-level function that raises errors when a problem is
@@ -99,7 +100,7 @@ async def send_trajectory(
     traj = Trajectory(fps, trajectories)
 
     if use_sync_line:
-        if not fps.ieb or fps.ieb.disabled:
+        if not isinstance(fps.ieb, IEB) or fps.ieb.disabled:
             raise TrajectoryError("IEB is not connected. Cannot use SYNC line.")
         if (await fps.ieb.get_device("sync").read())[0] == "closed":
             raise TrajectoryError("The SYNC line is on high.")
@@ -108,19 +109,19 @@ async def send_trajectory(
     await traj.send()
 
     if traj.failed:
-        raise TrajectoryError("something went wrong sending the trajectory.")
+        raise TrajectoryError("Something went wrong sending the trajectory.")
 
-    log.info(f"trajectory successfully sent in {traj.data_send_time:1f} seconds.")
-    log.info(f"expected time to complete trajectory: {traj.move_time:.2f} seconds.")
+    log.info(f"Trajectory successfully sent in {traj.data_send_time:1f} seconds.")
+    log.info(f"Expected time to complete trajectory: {traj.move_time:.2f} seconds.")
 
     log.info("starting trajectory ...")
     result = await traj.start(use_sync_line=use_sync_line)
     if traj.failed or not result:
-        raise TrajectoryError("something went wrong starting the trajectory.")
+        raise TrajectoryError("Something went wrong starting the trajectory.")
 
-    log.info("all positioners have successfully reached their positions.")
+    log.info("All positioners have successfully reached their positions.")
 
-    return True
+    return traj
 
 
 class Trajectory(object):
@@ -176,7 +177,7 @@ class Trajectory(object):
             raise FPSLockedError("FPS is locked. Cannot send trajectories.")
 
         if self.fps.moving:
-            raise TrajectoryError("the FPS is moving. Cannot send new trajectory.")
+            raise TrajectoryError("The FPS is moving. Cannot send new trajectory.")
 
         if isinstance(trajectories, (str, pathlib.Path)):
             self.trajectories = cast(TrajectoryDataType, read_yaml_file(trajectories))
@@ -207,19 +208,19 @@ class Trajectory(object):
             raise TrajectoryError("trajectory is empty.")
 
         if len(self.trajectories) != len(numpy.unique(list(self.trajectories.keys()))):
-            raise TrajectoryError("duplicate positioner trajectories.")
+            raise TrajectoryError("Duplicate positioner trajectories.")
 
         for pid in self.trajectories:
             trajectory = self.trajectories[pid]
 
             if "alpha" not in trajectory or "beta" not in trajectory:
-                raise TrajectoryError(f"positioner {pid} missing alpha or beta data.")
+                raise TrajectoryError(f"Positioner {pid} missing alpha or beta data.")
 
             for arm in ["alpha", "beta"]:
                 data = numpy.array(list(zip(*trajectory[arm]))[0])
 
                 if numpy.any(data > 360) or numpy.any(data < 0):
-                    raise TrajectoryError(f"positioner {pid} has points out of range.")
+                    raise TrajectoryError(f"Positioner {pid} has points out of range.")
 
                 if arm == "beta":
                     if config.get("safe_mode", False):
@@ -229,7 +230,7 @@ class Trajectory(object):
                             min_beta = config["safe_mode"]["min_beta"]
                         if numpy.any(data < min_beta):
                             raise TrajectoryError(
-                                f"positioner {pid}: safe mode is "
+                                f"Oositioner {pid}: safe mode is "
                                 f"on and beta < {min_beta}."
                             )
 
@@ -238,7 +239,7 @@ class Trajectory(object):
 
         self.move_time = 0.0
 
-        await self.abort_trajectory()
+        await self.fps.stop_trajectory()
 
         if not await self.fps.update_status(
             positioner_ids=list(self.trajectories.keys()),
@@ -290,7 +291,7 @@ class Trajectory(object):
         new_traj_cmds = [
             self.fps.send_command(
                 "SEND_NEW_TRAJECTORY",
-                positioner_id=pos_id,
+                positioner_ids=pos_id,
                 n_alpha=self.n_points[pos_id][0],
                 n_beta=self.n_points[pos_id][1],
             )
@@ -325,7 +326,7 @@ class Trajectory(object):
                     data_cmds.append(
                         self.fps.send_command(
                             "SEND_TRAJECTORY_DATA",
-                            positioner_id=pos_id,
+                            positioner_ids=pos_id,
                             positions=arm_chunk,
                         )
                     )
@@ -336,12 +337,13 @@ class Trajectory(object):
                     if cmd.status.failed or cmd.status.timed_out:
                         self.failed = True
                         raise TrajectoryError(
-                            "at least one SEND_TRAJECTORY_COMMAND failed."
+                            "At least one SEND_TRAJECTORY_COMMAND failed."
                         )
 
         # Finalise the trajectories
-        end_traj_cmds = await self.fps.send_to_all(
-            "TRAJECTORY_DATA_END", positioners=list(self.trajectories.keys())
+        end_traj_cmds = await self.fps.send_command(
+            "TRAJECTORY_DATA_END",
+            positioner_ids=list(self.trajectories.keys()),
         )
 
         for cmd in end_traj_cmds:
@@ -368,19 +370,18 @@ class Trajectory(object):
         """Starts the trajectory."""
 
         if not self._ready_to_start or self.failed:
-            raise TrajectoryError("the trajectory has not been sent.")
-
-        if use_sync_line:
-            if not self.fps.ieb or self.fps.ieb.disabled:
-                raise TrajectoryError("IEB is not connected. Cannot use SYNC line.")
-            if (await self.fps.ieb.get_device("sync").read())[0] == "closed":
-                raise TrajectoryError("The SYNC line is on high.")
+            raise TrajectoryError("The trajectory has not been sent.")
 
         for positioner_id in list(self.trajectories.keys()):
             self.fps[positioner_id].move_time = self.move_time
 
         if use_sync_line:
 
+            if not isinstance(self.fps.ieb, IEB) or self.fps.ieb.disabled:
+                raise TrajectoryError("IEB is not connected. Cannot use SYNC line.")
+            if (await self.fps.ieb.get_device("sync").read())[0] == "closed":
+                raise TrajectoryError("The SYNC line is on high.")
+
             sync = self.fps.ieb.get_device("sync")
             assert isinstance(sync, drift.Relay)
 
@@ -397,7 +398,7 @@ class Trajectory(object):
             # Start trajectories
             command = await self.fps.send_command(
                 "START_TRAJECTORY",
-                positioner_id=0,
+                positioner_ids=0,
                 timeout=1,
                 n_positioners=len(self.trajectories),
             )
@@ -407,50 +408,57 @@ class Trajectory(object):
                 self.failed = True
                 raise TrajectoryError("START_TRAJECTORY failed")
 
+        restart_pollers = True if self.fps.pollers.running else False
+        await self.fps.pollers.stop()
+
+        start_time = time.time()
+
         try:
-            await self.fps.pollers.set_delay(1)
-            use_pollers = True
-        except Exception as ee:
-            use_pollers = False
-            log.error(f"failed setting poller delay: {ee}.")
+            assert self.move_time is not None, "move_time not set."
 
-        assert self.move_time is not None, "move_time not set."
+            while True:
 
-        # Wait approximate time before starting to poll for status
-        await asyncio.sleep(0.95 * self.move_time)
+                await asyncio.sleep(1)
 
-        remaining_time = self.move_time - 0.95 * self.move_time
+                if self.fps.locked:
+                    raise TrajectoryError("The FPS got locked during the trajectory.")
 
-        # Wait until all positioners have completed.
-        wait_status = [
-            self.fps.positioners[pos_id].wait_for_status(
-                self.fps.positioners[pos_id].flags.DISPLACEMENT_COMPLETED,
-                timeout=remaining_time + 3,
-                delay=0.1,
-            )
-            for pos_id in self.trajectories
-        ]
-        results = await asyncio.gather(*wait_status)
+                await self.fps.update_status()
 
-        if not all(results):
-            if use_pollers:
-                await self.fps.pollers.set_delay()
+                if all(
+                    [
+                        maskbits.PositionerStatus.DISPLACEMENT_COMPLETED
+                        in self.fps[pid].status
+                        for pid in self.trajectories
+                    ]
+                ):
+                    self.failed = False
+                    break
+
+            elapsed = time.time() - start_time
+            if elapsed > (self.move_time + 3):
+                raise TrajectoryError("Some positioners did not complete the move.")
+
+        except BaseException:
             self.failed = True
-            raise TrajectoryError("some positioners did not complete the move.")
+            raise
 
-        # Restore default polling time
-        if use_pollers:
-            await self.fps.pollers.set_delay()
+        finally:
+            if restart_pollers:
+                self.fps.pollers.start()
 
         return True
 
     async def abort_trajectory(self):
         """Aborts the trajectory transmission."""
 
-        if not await self.fps.send_to_all(
-            "TRAJECTORY_TRANSMISSION_ABORT", positioners=list(self.trajectories.keys())
-        ):
-            raise TrajectoryError("cannot abort trajectory transmission.")
+        cmd = await self.fps.send_command(
+            "TRAJECTORY_TRANSMISSION_ABORT",
+            positioner_ids=list(self.trajectories.keys()),
+        )
+
+        if cmd.status.failed:
+            raise TrajectoryError("Cannot abort trajectory transmission.")
 
 
 class SendNewTrajectory(Command):
@@ -460,7 +468,7 @@ class SendNewTrajectory(Command):
     broadcastable = False
     move_command = True
 
-    def __init__(self, n_alpha, n_beta, **kwargs):
+    def __init__(self, positioner_ids, n_alpha, n_beta, **kwargs):
 
         alpha_positions = int(n_alpha)
         beta_positions = int(n_beta)
@@ -470,7 +478,7 @@ class SendNewTrajectory(Command):
         data = int_to_bytes(alpha_positions) + int_to_bytes(beta_positions)
         kwargs["data"] = data
 
-        super().__init__(**kwargs)
+        super().__init__(positioner_ids, **kwargs)
 
 
 class SendTrajectoryData(Command):
@@ -494,7 +502,7 @@ class SendTrajectoryData(Command):
     broadcastable = False
     move_command = True
 
-    def __init__(self, positions, **kwargs):
+    def __init__(self, positioner_ids, positions, **kwargs):
 
         positions = numpy.array(positions).astype(numpy.float64)
 
@@ -509,7 +517,7 @@ class SendTrajectoryData(Command):
 
         kwargs["data"] = data
 
-        super().__init__(**kwargs)
+        super().__init__(positioner_ids, **kwargs)
 
 
 class TrajectoryDataEnd(Command):
