@@ -6,14 +6,10 @@
 # @Filename: ieb.py
 # @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
 
-import warnings
-
 from typing import Any, Dict
 
 from drift import Drift, DriftError
 
-from jaeger.exceptions import JaegerUserWarning
-
 
 __all__ = ["IEB"]
 
@@ -56,10 +52,7 @@ class IEB(Drift):
             await Drift.__aenter__(self)
         except DriftError:
             self.disabled = True
-            warnings.warn(
-                "Failed connecting to the IEB. Disabling it.",
-                JaegerUserWarning,
-            )
+            raise DriftError("Failed connecting to the IEB. Disabling it.")
 
     async def __aexit__(self, *args):
 
