@@ -11,6 +11,7 @@ from __future__ import annotations
 import asyncio
 import logging
 from distutils.version import StrictVersion
+from time import time
 
 from typing import List, Optional, Tuple, cast
 
@@ -18,6 +19,7 @@ import numpy.testing
 
 import jaeger
 from jaeger import config, log, maskbits
+from jaeger.can import JaegerCAN
 from jaeger.commands import CommandID
 from jaeger.exceptions import JaegerError, PositionerError
 from jaeger.utils import StatusMixIn, bytes_to_int
@@ -143,6 +145,23 @@ class Positioner(StatusMixIn):
 
         log.log(level, f"Positioner {self.positioner_id}: {message}")
 
+    def get_bus(self) -> Tuple[int, int | None]:
+        """Returns the interface index and bus."""
+
+        if self.fps is None:
+            raise PositionerError("FPS is not defined.")
+
+        if not isinstance(self.fps.can, JaegerCAN):
+            raise PositionerError("CAN bus not found.")
+
+        if not self.fps.can.multibus:
+            return (0, None)
+
+        iface, bus = self.fps.positioner_to_bus[self.positioner_id]
+        iface_idx = self.fps.can.interfaces.index(iface)
+
+        return (iface_idx, bus)
+
     async def send_command(self, command, error: Optional[str] = None, **kwargs):
         """Sends and awaits a command to the FPS for this positioner."""
 
@@ -150,7 +169,9 @@ class Positioner(StatusMixIn):
             raise PositionerError("FPS is not set.")
 
         command = await self.fps.send_command(
-            command, positioner_id=self.positioner_id, **kwargs
+            command,
+            positioner_ids=self.positioner_id,
+            **kwargs,
         )
 
         if error and (command.status.failed or command.status.timed_out):
@@ -170,7 +191,6 @@ class Positioner(StatusMixIn):
             command = await self.send_command(
                 CommandID.GET_ACTUAL_POSITION,
                 timeout=timeout,
-                override=True,
             )
 
             if command.status.failed:
@@ -178,7 +198,7 @@ class Positioner(StatusMixIn):
                 raise PositionerError("failed updating position")
 
             try:
-                position = command.get_positions()
+                position = command.get_positions()[self.positioner_id]
             except ValueError:
                 raise PositionerError("cannot parse current position.")
 
@@ -193,6 +213,8 @@ class Positioner(StatusMixIn):
     ):
         """Updates the status of the positioner."""
 
+        assert self.fps, "FPS is not set."
+
         # Need to update the firmware to make sure we get the right flags.
         await self.update_firmware_version()
 
@@ -217,12 +239,12 @@ class Positioner(StatusMixIn):
         else:
             self.flags = maskbits.BootloaderStatus
 
-        self.status = self.flags(status)
+        self.status = self.flags(int(status))
 
         # Checks if the positioner is collided. If so, locks the FPS.
-        if not self.is_bootloader() and self.collision and not self.fps.locked:
-            await self.fps.lock()
-            raise PositionerError("collision detected. Locking the FPS.")
+        # if not self.is_bootloader() and self.collision and not self.fps.locked:
+        #     await self.fps.lock()
+        #     raise PositionerError("collision detected. Locking the FPS.")
 
         return True
 
@@ -275,7 +297,7 @@ class Positioner(StatusMixIn):
 
                 await asyncio.sleep(delay)
 
-        wait_for_status = [self.flags(ss) for ss in status]
+        wait_for_status = [self.flags(int(ss)) for ss in status]
 
         try:
             await asyncio.wait_for(status_waiter(wait_for_status), timeout)
@@ -284,7 +306,7 @@ class Positioner(StatusMixIn):
 
         return True
 
-    async def initialise(self):
+    async def initialise(self, disable_precise_moves=False):
         """Initialises the position watcher."""
 
         # Resets all.
@@ -309,6 +331,11 @@ class Positioner(StatusMixIn):
             beta=config["positioner"]["motor_speed"],
         )
 
+        if StrictVersion(self.firmware) < StrictVersion("04.01.17"):
+            self._log("Disabling precise moves requires >=04.01.17", logging.ERROR)
+        else:
+            await self.set_precise_move(mode=not disable_precise_moves)
+
         self._log("initialisation complete.")
 
         return True
@@ -321,7 +348,7 @@ class Positioner(StatusMixIn):
             error="failed retrieving firmware version.",
         )
 
-        self.firmware = command.get_firmware()
+        self.firmware = command.get_firmware(positioner_id=self.positioner_id)
         self.flags = self.get_positioner_flags()
 
         self._log(f"firmware {self.firmware}")
@@ -401,6 +428,33 @@ class Positioner(StatusMixIn):
 
         return speed_command
 
+    async def set_precise_move(self, mode, alpha=True, beta=True):
+        """Switches the precise moves on alpha and beta."""
+
+        cmds = []
+
+        if not alpha and not beta:
+            raise PositionerError("positioneralpha or beta need to be True.")
+
+        if alpha:
+            if mode is True:
+                cmds.append(self.send_command(CommandID.SWITCH_ON_PRECISE_MOVE_ALPHA))
+            else:
+                cmds.append(self.send_command(CommandID.SWITCH_OFF_PRECISE_MOVE_ALPHA))
+
+        if beta:
+            if mode is True:
+                cmds.append(self.send_command(CommandID.SWITCH_ON_PRECISE_MOVE_BETA))
+            else:
+                cmds.append(self.send_command(CommandID.SWITCH_OFF_PRECISE_MOVE_BETA))
+
+        cmds = await asyncio.gather(*cmds)
+
+        if any([cmd.status != maskbits.CommandStatus.DONE for cmd in cmds]):
+            raise PositionerError("failed switching precise moves.")
+
+        return True
+
     def _can_move(self):
         """Returns `True` if the positioner can be moved."""
 
@@ -509,6 +563,12 @@ class Positioner(StatusMixIn):
         if None in self.speed:
             raise PositionerError("speed has not been set.")
 
+        # Update position
+        await self.update_position()
+        assert (
+            self.alpha is not None and self.beta is not None
+        ), "alpha or beta positions unknown."
+
         # Check if safe mode is enabled
         if "safe_mode" in config and config["safe_mode"] is not False:
             if isinstance(config["safe_mode"], bool):
@@ -540,7 +600,7 @@ class Positioner(StatusMixIn):
 
             # Sleeps for the time the firmware believes it's going to take
             # to get to the desired position.
-            alpha_time, beta_time = goto_command.get_move_time()
+            alpha_time, beta_time = goto_command.get_move_time()[0]
 
             # Update status as soon as we start moving. This clears any
             # possible DISPLACEMENT_COMPLETED.
@@ -574,17 +634,28 @@ class Positioner(StatusMixIn):
 
             self._log(f"the move will take {self.move_time:.2f} seconds", logging.INFO)
 
-            await asyncio.sleep(self.move_time)
+            assert self.fps
 
-            # Blocks until we're sure both arms at at the position.
-            result = await self.wait_for_status(
-                self.flags.DISPLACEMENT_COMPLETED, delay=0.1, timeout=3
-            )
+            while True:
+                await asyncio.sleep(0.2)
 
-            if result is False:
-                raise PositionerError("failed to reach commanded position.")
+                if self.fps.locked:
+                    # No need to be very verbose about this.
+                    self._log(
+                        "Cancelling goto because the FPS got locked",
+                        logging.DEBUG,
+                    )
+                    result = False
+                    break
 
-            self._log("position reached.", logging.INFO)
+                elapsed = time() - goto_command.start_time
+                if elapsed > self.move_time + 3:
+                    raise PositionerError("failed to reach commanded position.")
+
+                await self.update_status()
+                if maskbits.PositionerStatus.DISPLACEMENT_COMPLETED in self.status:
+                    result = True
+                    break
 
         except BaseException:
             raise
@@ -593,7 +664,10 @@ class Positioner(StatusMixIn):
             if self.speed != original_speed:
                 await self.set_speed(*original_speed, force=force)
 
-        return True
+        if result is True:
+            self._log("position reached.", logging.INFO)
+
+        return result
 
     async def home(self):
         """Homes the positioner.
